from solaredge2mqtt.models.base import Component, EnumModel
from solaredge2mqtt.models.forecast import ForecastPeriod, ForecastQuery
from solaredge2mqtt.models.historic import (
    HistoricEnergy,
    HistoricMoney,
    HistoricPeriod,
    HistoricQuery,
)
from solaredge2mqtt.models.modbus import SunSpecBattery, SunSpecInverter, SunSpecMeter
from solaredge2mqtt.models.monitoring import (
    LogicalInfo,
    LogicalInverter,
    LogicalModule,
    LogicalString,
)
from solaredge2mqtt.models.powerflow import Powerflow
from solaredge2mqtt.models.wallbox import WallboxAPI
from solaredge2mqtt.models.weather import (
    OpenWeatherMapBaseData,
    OpenWeatherMapCurrentData,
    OpenWeatherMapForecastData,
    OpenWeatherMapOneCall,
    OpenWeatherMapOneCallTimemachine,
)



from solaredge2mqtt.models.base import EnumModel


class ForecastQuery(EnumModel):
    ACTUAL = "actual_unit"
    NEXT = "forecast_unit"

    def __init__(self, query: str) -> None:
        self._query: str = query

    @property
    def query(self) -> str:
        return self._query


class ForecastPeriod(EnumModel):
    TODAY = "today", "1d", ForecastQuery.ACTUAL
    TOMORROW = "tomorrow", "1d", ForecastQuery.NEXT
    CURRENT_HOUR = "current_hour", "1h", ForecastQuery.ACTUAL
    NEXT_HOUR = "next_hour", "1h", ForecastQuery.NEXT

    def __init__(self, topic: str, unit: str, query: ForecastQuery) -> None:
        self._topic: str = topic
        self._unit: str = unit
        self._query: ForecastQuery = query

    @property
    def topic(self) -> str:
        return self._topic

    @property
    def unit(self) -> str:
        return self._unit

    @property
    def query(self) -> ForecastQuery:
        return self._query



from __future__ import annotations

from pydantic import BaseModel


class LogicalInfo(BaseModel):
    id: str
    serialnumber: str | None
    name: str
    type: str

    @staticmethod
    def map(data: dict[str, str | int]) -> dict[str, str]:
        return {
            "id": str(data["id"]),
            "serialnumber": data["serialNumber"],
            "name": data["name"],
            "type": data["type"],
        }


class LogicalInverter(BaseModel):
    info: LogicalInfo
    energy: float | None
    strings: list[LogicalString] = []


class LogicalString(BaseModel):
    info: LogicalInfo
    energy: float | None
    modules: list[LogicalModule] = []


class LogicalModule(BaseModel):
    info: LogicalInfo
    energy: float | None



from solaredge2mqtt.exceptions import ConfigurationException, InvalidDataException
from solaredge2mqtt.logging import logger
from solaredge2mqtt.models import (
    HistoricEnergy,
    HistoricPeriod,
    HistoricQuery,
    Powerflow,
)
from solaredge2mqtt.mqtt import MQTTClient
from solaredge2mqtt.service.influxdb import InfluxDB
from solaredge2mqtt.service.modbus import Modbus
from solaredge2mqtt.service.wallbox import WallboxClient
from solaredge2mqtt.settings import ServiceSettings


class BaseLoops:
    def __init__(
        self,
        settings: ServiceSettings,
        mqtt: MQTTClient,
        influxdb: InfluxDB | None = None,
    ):
        self.settings = settings

        self.mqtt = mqtt

        self.influxdb = influxdb

        self.modbus = Modbus(self.settings.modbus)

        self.wallbox = (
            WallboxClient(self.settings.wallbox)
            if self.settings.is_wallbox_configured
            else None
        )

    async def powerflow_loop(self) -> None:
        inverter_data, meters_data, batteries_data = await self.modbus.loop()

        if any(data is None for data in [inverter_data, meters_data, batteries_data]):
            raise InvalidDataException("Invalid modbus data")

        for battery in batteries_data.values():
            if not battery.is_valid:
                logger.debug(battery)
                raise InvalidDataException("Invalid battery data")

        evcharger = 0
        wallbox_data = None
        if self.settings.is_wallbox_configured:
            try:
                wallbox_data = await self.wallbox.loop()
                logger.trace(
                    "Wallbox: {wallbox_data.power} W", wallbox_data=wallbox_data
                )
                evcharger = wallbox_data.power
            except ConfigurationException as ex:
                logger.warning(f"{ex.component}: {ex.message}")

        evcharger = 0

        powerflow = Powerflow(inverter_data, meters_data, batteries_data, evcharger)
        if not powerflow.is_valid:
            logger.info(powerflow)
            raise InvalidDataException("Invalid powerflow data")

        if Powerflow.is_not_valid_with_last(powerflow):
            logger.debug(powerflow)
            raise InvalidDataException("Value change not valid, skipping this loop")

        logger.debug(powerflow)
        logger.info(
            "Powerflow: PV {pv_production} W, Inverter {inverter.power} W, "
            + "House {consumer.house} W, Grid {grid.power} W, "
            + "Battery {battery.power} W, Wallbox {consumer.evcharger} W",
            pv_production=powerflow.pv_production,
            inverter=powerflow.inverter,
            consumer=powerflow.consumer,
            grid=powerflow.grid,
            battery=powerflow.battery,
        )

        await self.mqtt.publish_to(inverter_data.mqtt_topic, inverter_data)
        for key, component in {**meters_data, **batteries_data}.items():
            await self.mqtt.publish_to(
                f"{component.mqtt_topic}/{key.lower()}", component
            )

        if wallbox_data is not None:
            await self.mqtt.publish_to(wallbox_data.mqtt_topic, wallbox_data)

        await self.mqtt.publish_to("powerflow", powerflow)

        if self.influxdb is not None:
            points = [powerflow.prepare_point()]

            for battery in batteries_data.values():
                points.append(battery.prepare_point())

            self.influxdb.write_points(points)

    async def energy_loop(self):
        for period in HistoricPeriod:
            record = self.influxdb.query_timeunit(period, "energy")
            if record is None:
                if period.query == HistoricQuery.LAST:
                    logger.info(
                        "No data found for {period}, skipping this loop", period=period
                    )
                else:
                    raise InvalidDataException(f"No energy data for {period}")

                continue

            energy = HistoricEnergy(record, period)

            logger.info(
                "Read from influxdb {period} energy: {energy.pv_production} kWh",
                period=period,
                energy=energy,
            )

            await self.mqtt.publish_to(f"energy/{period.topic}", energy)



"""
    This module, service.py, is part of the SolarEdge2MQTT service, which reads data 
    from a SolarEdge inverter and publishes it to an MQTT broker. It uses the asyncio 
    library for asynchronous I/O and the aiomqtt library for MQTT communication. 
    The module also includes a run function to initialize and start the service.
"""

import asyncio as aio
import signal
from time import time
from typing import Callable

from aiomqtt import MqttError

from solaredge2mqtt import __version__
from solaredge2mqtt.exceptions import ConfigurationException, InvalidDataException
from solaredge2mqtt.logging import initialize_logging, logger
from solaredge2mqtt.mqtt import MQTTClient
from solaredge2mqtt.service.base import BaseLoops
from solaredge2mqtt.service.forecast import Forecast
from solaredge2mqtt.service.influxdb import InfluxDB
from solaredge2mqtt.service.monitoring import MonitoringSite
from solaredge2mqtt.service.weather import WeatherClient
from solaredge2mqtt.settings import service_settings


def run():
    try:
        service = Service()
        loop = aio.get_event_loop()
        loop.add_signal_handler(signal.SIGINT, service.cancel)
        loop.add_signal_handler(signal.SIGTERM, service.cancel)
        loop.run_until_complete(service.main_loop())
        loop.close()
    except ConfigurationException:
        logger.error("Configuration error")


class Service:
    def __init__(self):
        self.settings = service_settings()
        initialize_logging(self.settings.logging_level)
        logger.debug(self.settings)

        self.mqtt = MQTTClient(self.settings.mqtt)

        self.cancel_request = aio.Event()
        self.loops: set[aio.Task] = set()

        self.influxdb: InfluxDB | None = (
            InfluxDB(self.settings.influxdb, self.settings.prices)
            if self.settings.is_influxdb_configured
            else None
        )

        self.basics = BaseLoops(self.settings, self.mqtt, self.influxdb)

        self.monitoring: MonitoringSite | None = (
            MonitoringSite(self.settings.monitoring, self.mqtt)
            if self.settings.is_monitoring_configured
            else None
        )

        self.weather: WeatherClient | None = (
            WeatherClient(self.settings, self.mqtt)
            if self.settings.is_weather_configured
            else None
        )

        self.forecast: Forecast | None = (
            Forecast(
                self.settings.forecast,
                self.settings.location,
                self.mqtt,
                self.influxdb,
                self.weather,
            )
            if self.settings.is_forecast_configured
            else None
        )

    def cancel(self):
        logger.info("Stopping SolarEdge2MQTT service...")
        self.cancel_request.set()
        for loop in self.loops:
            loop.cancel()

    async def main_loop(self):
        logger.info("Starting SolarEdge2MQTT service...")
        logger.info("Version: {version}", version=__version__)
        logger.debug(self.settings)
        logger.info("Timezone: {timezone}", timezone=self.settings.influxdb.timezone)

        if self.settings.is_influxdb_configured:
            self.influxdb.initialize_buckets()
            # self.influxdb.initialize_task()

        while not self.cancel_request.is_set():
            try:
                async with self.mqtt:
                    await self.mqtt.publish_status_online()

                    self.schedule_loop(
                        self.settings.interval, self.basics.powerflow_loop
                    )

                    self.schedule_influxdb_loops()

                    self.schedule_monitoring_loop()

                    await self.schedule_weather_loops()

                    await aio.gather(*self.loops)

                    await self.mqtt.publish_status_offline()

            except MqttError:
                logger.error("MQTT error, reconnecting in 5 seconds...")
                await aio.sleep(5)
            except aio.exceptions.CancelledError:
                logger.debug("Loops cancelled")
                return
            finally:
                for loop in self.loops:
                    loop.cancel()

    def schedule_monitoring_loop(self):
        if self.settings.is_monitoring_configured:
            self.monitoring.login()
            self.schedule_loop(300, self.monitoring.loop)

    def schedule_influxdb_loops(self):
        if self.settings.is_influxdb_configured:
            loop_handles = [self.influxdb.loop, self.basics.energy_loop]
            if self.settings.is_forecast_configured:
                loop_handles.append(self.forecast.training_loop)

            self.schedule_loop(600, loop_handles)

    async def schedule_weather_loops(self):
        if self.settings.is_weather_configured:
            self.schedule_loop(600, self.weather.loop)

            if self.settings.is_forecast_configured:
                self.schedule_loop(600, self.forecast.forecast_loop, True)

    def schedule_loop(
        self,
        interval_in_seconds: int,
        handles: Callable | list[Callable],
        delay_start: bool = False,
        args: list[any] = None,
    ):
        loop = aio.create_task(
            self.run_loop(interval_in_seconds, handles, delay_start, args)
        )
        self.loops.add(loop)
        loop.add_done_callback(self.loops.remove)

    async def run_loop(
        self,
        interval_in_seconds: int,
        handles: Callable | list[Callable],
        delay_start: bool = False,
        args: list[any] = None,
    ):
        if not isinstance(handles, list):
            handles = [handles]

        if delay_start:
            await aio.sleep(interval_in_seconds)

        while not self.cancel_request.is_set():
            execution_time = 0
            try:
                start_time = time()
                for handle in handles:
                    await handle(*args or [])
                execution_time = time() - start_time
            except InvalidDataException as error:
                logger.warning("{message}, skipping this loop", message=error.message)

            if execution_time < interval_in_seconds:
                await aio.sleep(interval_in_seconds - execution_time)
            else:
                await aio.sleep(interval_in_seconds)



from __future__ import annotations

import time
from asyncio import Event, to_thread
from datetime import datetime, timedelta, timezone

import ephem
from astral import LocationInfo
from astral.sun import azimuth, elevation, sun
from numpy import cos, percentile, pi, sin
from pandas import DataFrame, Series, to_datetime
from sklearn import clone
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.compose import ColumnTransformer
from sklearn.ensemble import HistGradientBoostingRegressor
from sklearn.inspection import permutation_importance
from sklearn.model_selection import GridSearchCV, TimeSeriesSplit, train_test_split
from sklearn.pipeline import Pipeline

from solaredge2mqtt.exceptions import InvalidDataException
from solaredge2mqtt.logging import logger
from solaredge2mqtt.models import EnumModel, OpenWeatherMapForecastData
from solaredge2mqtt.mqtt import MQTTClient
from solaredge2mqtt.service.influxdb import InfluxDB, Point
from solaredge2mqtt.service.weather import WeatherClient
from solaredge2mqtt.settings import ForecastSettings, LocationSettings


class ForecasterType(EnumModel):
    ENERGY = "energy"
    POWER = "power"

    def __init__(self, target_column: str) -> None:
        self._target_column: str = target_column

    @property
    def target_column(self) -> str:
        return self._target_column

    def prepare_value(self, value: float | int) -> float | int:
        if value <= 0:
            prepared = 0
        elif self.target_column == "energy":
            prepared = round(value / 1000, 3)
        else:
            prepared = int(round(value))

        return prepared


class Forecast:
    def __init__(
        self,
        settings: ForecastSettings,
        location: LocationSettings,
        mqtt: MQTTClient,
        influxdb: InfluxDB,
        weather: WeatherClient,
    ) -> None:
        self.settings = settings
        self.location = location
        self.mqtt = mqtt
        self.influxdb = influxdb
        self.weather = weather

        self.forecasters: dict[ForecasterType, Forecaster] = {
            typed: Forecaster(typed, location, settings.hyperparametertuning)
            for typed in ForecasterType
        }

        self.last_weather_forecast: list[OpenWeatherMapForecastData] | None = None

    async def training_loop(self):
        now = datetime.now().astimezone()
        rounded_minutes = (now.minute // 10) * 10
        last_hour = now.replace(minute=0, second=0, microsecond=0) - timedelta(hours=1)

        training_data = self.get_weather_training(last_hour)
        training_data["time"] = last_hour
        training_data = self.add_last_hour_pv_production(training_data)

        self.write_new_training_data_to_influxdb(training_data)

        if (
            rounded_minutes == 10
            or not self.forecasters[ForecasterType.ENERGY].is_trained
            or not self.forecasters[ForecasterType.POWER].is_trained
        ):
            await self.train()

    def get_weather_training(
        self, forecast_time: datetime
    ) -> dict[str, str | float | int | None]:
        if (
            self.last_weather_forecast is None
            or forecast_time.hour != self.last_weather_forecast[0].hour
        ):
            logger.info("Retrieve weather forecast for training of production forecast")
            self.last_weather_forecast = self.weather.get_weather().hourly

        return self.last_weather_forecast[0].model_dump_estimation_data()

    def add_last_hour_pv_production(
        self, trainings_data
    ) -> dict[str, str | float | int | None]:
        production_data = self.influxdb.query_first("production")

        if production_data is None:
            raise InvalidDataException("Missing production data for last 10 minutes")

        trainings_data[ForecasterType.POWER.target_column] = round(
            production_data[ForecasterType.POWER.target_column]
        )
        trainings_data[ForecasterType.ENERGY.target_column] = round(
            production_data[ForecasterType.ENERGY.target_column], 2
        )
        return trainings_data

    def write_new_training_data_to_influxdb(self, trainings_data):
        point = Point("forecast_training")
        for key, value in trainings_data.items():
            if isinstance(value, (int, float, str, bool)):
                point.field(key, value)

        point.time(trainings_data["time"].astimezone(timezone.utc))

        logger.info("Write new training data to influxdb")
        logger.debug(trainings_data)
        self.influxdb.write_point(point)

    async def train(self) -> None:
        data = await self.influxdb.query_dataframe("training_data")
        data["time"] = data["_time"].dt.tz_convert(self.location.timezone)
        await to_thread(self.training, data)

    def training(self, data: DataFrame) -> None:
        for forecaster in self.forecasters.values():
            forecaster.train(data)
            # self.replay_priHediction(data, forecaster)

    async def replay_prediction(self, data: DataFrame, forecaster: Forecaster) -> None:
        """FOR DEVELOPMENT PURPOSE ONLY"""
        logger.warning("Replaying prediction for development purpose")
        data = await forecaster.predict(data)
        data["time"] = data["_time"]
        periods = self._calculate_periods_prediction(data, forecaster.typed)
        self._write_periods_to_influxdb(periods, forecaster.typed)

    async def forecast_loop(self):
        if (
            not self.forecasters[ForecasterType.ENERGY].is_trained
            or not self.forecasters[ForecasterType.POWER].is_trained
        ):
            raise InvalidDataException("Forecast model is not trained yet")

        if self.last_weather_forecast is None:
            raise InvalidDataException("Missing weather forecast for forecast")

        estimation_data_list = [
            {
                "time": datetime(
                    year=weather_forecast.year,
                    month=weather_forecast.month,
                    day=weather_forecast.day,
                    hour=weather_forecast.hour,
                    minute=0,
                    second=0,
                    microsecond=0,
                ).astimezone(),
                **weather_forecast.model_dump_estimation_data(),
            }
            for weather_forecast in self.last_weather_forecast
        ]

        data = DataFrame(estimation_data_list)

        for typed, forecaster in self.forecasters.items():
            predicted_data = await forecaster.predict(data)

            periods = self._calculate_periods_prediction(predicted_data, typed)
            self._write_periods_to_influxdb(periods, typed)

            hours = self._calculate_hourly_prediction(periods, typed)
            logger.debug("Forecasted next hour: {hours}", hours=hours)
            await self._publish_hours_to_mqtt(hours, typed)

    def _calculate_periods_prediction(
        self, data: DataFrame, typed: ForecasterType
    ) -> DataFrame:
        data["year"] = data["time"].dt.year
        data["month"] = data["time"].dt.month
        data["day"] = data["time"].dt.day
        data["hour"] = data["time"].dt.hour

        return data[["time", "year", "month", "day", "hour", typed.target_column]]

    def _calculate_hourly_prediction(
        self, data: DataFrame, typed: ForecasterType
    ) -> DataFrame:

        if typed == ForecasterType.ENERGY:
            hourly = data.groupby(["year", "month", "day", "hour"], as_index=False)[
                ForecasterType.ENERGY.target_column
            ].sum()
        elif typed == ForecasterType.POWER:
            hourly = data.groupby(["year", "month", "day", "hour"], as_index=False)[
                ForecasterType.POWER.target_column
            ].mean()

        hourly["time"] = hourly.apply(
            lambda row: to_datetime(
                f"{int(row['year'])}-{int(row['month'])}-{int(row['day'])}T{int(row['hour'])}:00:00"
            ).tz_localize(self.location.timezone),
            axis=1,
        )

        return hourly[["time", "year", "month", "day", "hour", typed.target_column]]

    def _write_periods_to_influxdb(
        self, periods: DataFrame, typed: ForecasterType
    ) -> None:
        points = []
        for _, period in periods.iterrows():
            point = Point("forecast")
            point.field(typed.target_column, period[typed.target_column])
            utc_time = period["time"].astimezone(timezone.utc)
            point.time(utc_time)
            points.append(point)

        logger.info("Write forecast data to influxdb")
        self.influxdb.write_points(points)

    async def _publish_hours_to_mqtt(
        self, hours: DataFrame, typed: ForecasterType
    ) -> None:
        data = hours[["time", typed.target_column]]
        data = self._replace_time_with_iso(data)
        await self.mqtt.publish_to(
            f"forecast/{typed.target_column}/hourly", data.to_json(orient="records")
        )

    @staticmethod
    def _replace_time_with_iso(data: DataFrame) -> str:
        data = data.copy()
        data["time"] = data["time"].dt.strftime("%Y-%m-%dT%H:%M:%S%z")
        return data


class Forecaster:
    NUMERIC_FEATURES: list[str] = [
        "clouds",
        "dew_point",
        "feels_like",
        "humidity",
        "pop",
        "pressure",
        "rain",
        "temp",
        "uvi",
        "visibility",
        "wind_speed",
        "wind_gust",
    ]
    CATEGORICAL_FEATURES: list[str] = ["weather_id", "weather_main"]
    CYCLICAL_FEATURES: dict[str, int] = {
        "wind_deg": 360,
    }
    TARGET_FEATURES: dict[str, str] = {
        ForecasterType.ENERGY.target_column: "energy",
        ForecasterType.POWER.target_column: "power",
    }

    def __init__(
        self,
        typed: ForecasterType,
        location: LocationSettings,
        enable_hyperparameter_tuning: bool = False,
    ) -> None:
        self.typed: ForecasterType = typed
        self.location = location
        self.enable_hyperparameter_tuning = enable_hyperparameter_tuning
        self.model_pipeline: Pipeline = None
        self.training_completed: Event = Event()

    def train(self, data: DataFrame) -> None:
        data_count = len(data)
        if data_count < 60:
            raise InvalidDataException("Not enough data to train the model")

        logger.info(f"Training model {self.typed} with {data_count} data points")

        self.training_completed.clear()
        start_time = time.time()
        y_vector = data[self.typed.target_column]

        pipeline = self._prepare_model_pipeline(data.columns.to_list())

        if self.enable_hyperparameter_tuning:
            self.model_pipeline = self._hyperparametertuning(data, y_vector, pipeline)
        else:
            self.model_pipeline = pipeline

        self.model_pipeline.fit(data, y_vector)

        execution_time = time.time() - start_time
        self.training_completed.set()

        transformed_features = self.model_pipeline[
            "preprocessor"
        ].get_feature_names_out()
        logger.debug(
            "Transformed features ({count}): {features} ",
            count=len(transformed_features),
            features=", ".join(transformed_features),
        )

        selected_features = self.model_pipeline[
            "feature_selector"
        ].important_features_.tolist()
        logger.info(
            "Selected features ({count}): {features} ",
            count=len(selected_features),
            features=", ".join(selected_features),
        )

        logger.info(f"Training execution time: {execution_time:.2f} seconds")

    def _hyperparametertuning(self, data, y_vector, pipeline):
        param_grid = {
            "model__max_iter": [100, 200, 300],
            "model__max_depth": [None, 5, 10],
            "model__learning_rate": [0.01, 0.1],
        }

        grid_search = GridSearchCV(
            estimator=pipeline,
            param_grid=param_grid,
            cv=TimeSeriesSplit(n_splits=2),
            scoring="neg_mean_squared_error",
        )
        grid_search.fit(data, y_vector)

        logger.info(
            "Training with best parameters: {params}", params=grid_search.best_params_
        )
        logger.info("Training with best score: {score}", score=grid_search.best_score_)

        return clone(grid_search.best_estimator_)

    def _prepare_model_pipeline(self, x_vector_columns: list[str]) -> None:
        return Pipeline(
            steps=[
                ("preprocessor", self._prepare_preprocessor(x_vector_columns)),
                (
                    "feature_selector",
                    PFISelector(
                        estimator=HistGradientBoostingRegressor(
                            random_state=42, categorical_features="from_dtype"
                        )
                    ),
                ),
                (
                    "model",
                    HistGradientBoostingRegressor(
                        random_state=42, categorical_features="from_dtype"
                    ),
                ),
            ]
        )

    def _prepare_preprocessor(self, x_vector_columns: list[str]) -> ColumnTransformer:
        ct = ColumnTransformer(
            transformers=[
                (
                    "cyc",
                    CyclicalEncoder(**self.CYCLICAL_FEATURES),
                    self._extract_used_columns(
                        self.CYCLICAL_FEATURES, x_vector_columns
                    ),
                ),
                (
                    "num",
                    "passthrough",
                    self._extract_used_columns(self.NUMERIC_FEATURES, x_vector_columns),
                ),
                (
                    "time",
                    TimeEncoder(),
                    ["time"],
                ),
                ("sun", SunEncoder(self.location), ["time"]),
                (
                    "cat",
                    CategoricalEncoder(),
                    self._extract_used_columns(
                        self.CATEGORICAL_FEATURES, x_vector_columns
                    ),
                ),
            ],
            remainder="drop",
        )
        ct.set_output(transform="pandas")
        return ct

    @staticmethod
    def _extract_used_columns(
        typed_features: list[str] | dict[str, int], x_vector_columns: list[str]
    ) -> list[str]:
        return [col for col in typed_features if col in x_vector_columns]

    @property
    def is_trained(self) -> bool:
        return self.model_pipeline is not None

    async def predict(self, new_data: DataFrame) -> DataFrame:
        if self.model_pipeline is None:
            raise InvalidDataException("The model has not been trained yet.")

        data_for_prediction = new_data.copy()

        await self.training_completed.wait()

        predictions = self.model_pipeline.predict(data_for_prediction)
        data_for_prediction[self.typed.target_column] = predictions
        data_for_prediction[self.typed.target_column] = data_for_prediction[
            self.typed.target_column
        ].apply(self.typed.prepare_value)

        return data_for_prediction


class BaseEncoder(BaseEstimator, TransformerMixin):
    def __init__(self) -> None:
        self.features: list[str] | None = None
        self._feature_names_out: list[str] = []

    def fit(self, x_vector: DataFrame, *_) -> BaseEncoder:
        if not hasattr(x_vector, "columns"):
            raise AttributeError("x_vector has no columns")

        self.features = x_vector.columns.tolist()
        return self

    def _transform(self, x_vector: DataFrame) -> DataFrame:
        if self.features is None:
            raise AttributeError(f"Encoder {self.__class__} is not been fitted yet.")

        if not hasattr(x_vector, "columns"):
            raise AttributeError("x_vector has no columns")

        if not all(feature in self.features for feature in x_vector.columns):
            raise AttributeError(f"Columns {x_vector.columns} are not in the vector")

        if not all(feature in x_vector.columns for feature in self.features):
            raise AttributeError(f"Columns {self.features} are not in the vector")

        return x_vector

    def _save_feature_names_out(self, x_vector: DataFrame) -> DataFrame:
        self._feature_names_out = x_vector.columns.to_list()
        return x_vector

    def get_feature_names_out(self, *_) -> list[str]:
        return self._feature_names_out


class CategoricalEncoder(BaseEncoder):
    def transform(self, x_vector: DataFrame) -> DataFrame:
        x_vector = self._transform(x_vector).astype("category")
        return self._save_feature_names_out(x_vector)


class CyclicalEncoder(BaseEncoder):
    def __init__(self, **cycle_lengths: dict[str, int]) -> None:
        super().__init__()
        self.cycle_lengths: dict[str, int] = cycle_lengths

    def transform(self, x_vector: DataFrame) -> DataFrame:
        x_vector = self._transform(x_vector)
        for feature in self.features:
            cycle = self.cycle_lengths.get(feature, None)
            if not cycle:
                raise ValueError(f"Unknown cyclical feature {feature}")

            x_vector = self.transform_cycle_columns(
                x_vector, feature, x_vector[feature], cycle
            )
            x_vector.drop(feature, axis=1, inplace=True)

        return self._save_feature_names_out(x_vector)

    def get_params(self, deep=True) -> dict[str, int]:
        return self.cycle_lengths

    @staticmethod
    def transform_cycle_columns(
        x_vector: DataFrame, prefix: str, cycle_vector: DataFrame, cycle_length: float
    ) -> DataFrame:
        x_vector[f"{prefix}_cos"] = cos(2 * pi * cycle_vector / cycle_length)
        x_vector[f"{prefix}_sin"] = sin(2 * pi * cycle_vector / cycle_length)

        return x_vector


class TimeEncoder(BaseEncoder):
    def __init__(self) -> None:
        super().__init__()
        self.season_starts: dict[int, dict[str, datetime]] = {}

    def transform(self, x_vector: DataFrame) -> DataFrame:
        x_vector = self._transform(x_vector)
        for feature in self.features:
            x_vector = CyclicalEncoder.transform_cycle_columns(
                x_vector, f"{feature}_hour", x_vector[feature].dt.hour, 24
            )

            x_vector = CyclicalEncoder.transform_cycle_columns(
                x_vector, f"{feature}_month", x_vector[feature].dt.month, 12
            )

            x_vector[f"{feature}_dst"] = (
                x_vector[feature]
                .apply(lambda x: x.dst() != timedelta(0))
                .astype("category")
            )

            x_vector[f"{feature}_season"] = (
                x_vector[feature].apply(self._map_season).astype("category")
            )
            x_vector = CyclicalEncoder.transform_cycle_columns(
                x_vector,
                f"{feature}_day_of_year",
                x_vector[feature].dt.dayofyear,
                365.25,
            )

            x_vector.drop(feature, axis=1, inplace=True)

        logger.trace(x_vector.head(30))
        return self._save_feature_names_out(x_vector)

    def _map_season(self, date: datetime) -> str:
        year = date.year
        if year not in self.season_starts:
            equinox_mar = ephem.next_vernal_equinox(str(year))
            solstice_jun = ephem.next_summer_solstice(equinox_mar)
            equinox_sep = ephem.next_autumnal_equinox(solstice_jun)
            solstice_dec = ephem.next_winter_solstice(equinox_sep)

            self.season_starts[year] = {
                "spring": equinox_mar.datetime().astimezone(),
                "summer": solstice_jun.datetime().astimezone(),
                "autumn": equinox_sep.datetime().astimezone(),
                "winter": solstice_dec.datetime().astimezone(),
            }

        starts = self.season_starts[year]

        season = "winter"

        if date >= starts["spring"]:
            if date < starts["summer"]:
                season = "spring"
            elif date < starts["autumn"]:
                season = "summer"
            elif date < starts["winter"]:
                season = "autumn"

        return season


class SunEncoder(BaseEncoder):
    def __init__(self, location: LocationSettings) -> None:
        super().__init__()
        self.location = location
        self._location = LocationInfo(
            "name",
            "region",
            timezone=location.timezone,
            latitude=location.latitude,
            longitude=location.longitude,
        )

    def transform(self, x_vector: DataFrame) -> DataFrame:
        x_vector = self._transform(x_vector)

        for feature in self.features:
            time_key = f"{feature}_time"

            x_vector[time_key] = x_vector[feature].apply(
                lambda x: x + timedelta(minutes=30)
            )

            x_vector[f"{feature}_elevation"] = x_vector[time_key].apply(
                lambda x: elevation(self._location.observer, x)
            )

            x_vector = CyclicalEncoder.transform_cycle_columns(
                x_vector,
                f"{feature}_azimuth",
                x_vector[time_key].apply(lambda x: azimuth(self._location.observer, x)),
                360,
            )

            x_vector[
                [
                    f"{feature}_daylight",
                    f"{feature}_delta_sunrise",
                    f"{feature}_delta_sunset",
                ]
            ] = x_vector[time_key].apply(self.daylight_info)

            x_vector.drop([feature, time_key], axis=1, inplace=True)

        return self._save_feature_names_out(x_vector)

    def daylight_info(self, row_time: datetime) -> Series:
        s = sun(self._location.observer, row_time)
        daylight = (s["sunset"] - s["sunrise"]).total_seconds() / 3600

        delta_sunrise = (row_time - s["sunrise"]).total_seconds() / 3600
        delta_sunset = (s["sunset"] - row_time).total_seconds() / 3600

        return Series([daylight, delta_sunrise, delta_sunset])


class PFISelector(BaseEstimator, TransformerMixin):
    def __init__(self, estimator, n_repeats=10):
        self.estimator = estimator
        self.n_repeats = n_repeats

        self.estimator_ = None
        self.feature_importances_ = None
        self.important_indices_ = None
        self.important_features_ = None

    def fit(self, x_vector: DataFrame, y_vector=None) -> PFISelector:
        x_train, x_test, y_train, y_test = train_test_split(
            x_vector, y_vector, test_size=0.1, random_state=42
        )
        self.estimator_ = self.estimator.fit(x_train, y_train)
        results = permutation_importance(
            self.estimator_,
            x_test,
            y_test,
            n_repeats=self.n_repeats,
            random_state=42,
            n_jobs=-1,
        )

        self.feature_importances_ = results.importances_mean

        threshold_value = percentile(self.feature_importances_, 75)

        self.important_indices_ = self.feature_importances_ > threshold_value
        self.important_features_ = x_vector.columns[self.important_indices_]
        return self

    def transform(self, x_vector: DataFrame) -> DataFrame:
        return x_vector[self.important_features_]

    def get_support(self, *_) -> list[bool]:
        return self.important_indices_.to_list()



import time

import jwt
import urllib3
from pydantic import BaseModel, Field
from requests.exceptions import HTTPError

from solaredge2mqtt.exceptions import ConfigurationException, InvalidDataException
from solaredge2mqtt.logging import logger
from solaredge2mqtt.models import WallboxAPI
from solaredge2mqtt.settings import WallboxSettings
from solaredge2mqtt.service.http import HTTPClient

urllib3.disable_warnings()

LOGIN_URL = "https://{host}:8443/v2/jwt/login"
REFRESH_URL = "https://{host}:8443/v2/jwt/refresh"
WALLBOX_URL = "https://{host}:8443/v2/wallboxes/{serial}"


class AuthorizationTokens(BaseModel):
    access_token: str | None = Field(None, alias="accessToken")
    refresh_token: str | None = Field(None, alias="refreshToken")

    @property
    def access_token_expires(self) -> int:
        payload = jwt.decode(self.access_token, options={"verify_signature": False})
        logger.debug("Access token expires at: {expires}", expires=payload["exp"])
        return payload["exp"]

    @property
    def refresh_token_expires(self) -> int:
        payload = jwt.decode(self.refresh_token, options={"verify_signature": False})
        logger.debug("Refresh token expires at: {expires}", expires=payload["exp"])
        return payload["exp"]


class WallboxClient(HTTPClient):
    def __init__(self, settings: WallboxSettings):
        super().__init__("Wallbox API")
        self.settings = settings

        logger.info(
            "Using Wallbox charger: {host}",
            host=settings.host,
        )

        self.authorization: AuthorizationTokens | None = None

    async def loop(self) -> WallboxAPI | None:
        wallbox = None
        try:
            self._get_access()

            response = self._get(
                WALLBOX_URL.format(
                    host=self.settings.host, serial=self.settings.serial
                ),
                headers={"Authorization": f"Bearer {self.authorization.access_token}"},
                verify=False,
                login=self.login,
            )

            if response is None:
                raise InvalidDataException("Invalid Wallbox data")

            wallbox = WallboxAPI(response)
            logger.info(
                f"Wallbox: {wallbox.state}, {wallbox.power / 1000} kW",
                wallbox=wallbox,
            )
        except HTTPError as error:
            raise InvalidDataException(f"Cannot read Wallbox data: {error}") from error

        return wallbox

    def _get_access(self) -> None:
        current_timestamp = int(time.time())

        if self.authorization is None:
            self.login()
        elif self.authorization.access_token_expires < current_timestamp + 60:
            # Token is about to expire within 60 seconds
            if self.authorization.refresh_token_expires < current_timestamp + 60:
                # Refresh token is about to expire within 60 seconds as well new login
                self.login()
            else:
                self._refresh_token()
        else:
            logger.debug("Wallbox access token still valid")

    def login(self):
        try:
            logger.info("Logging in to Wallbox charger...")
            self.authorization = None
            response = self._post(
                LOGIN_URL.format(host=self.settings.host),
                {
                    "password": self.settings.password.get_secret_value(),
                    "username": "admin",
                },
                verify=False,
            )
            logger.trace(response)

            if response is None:
                raise ConfigurationException("wallbox", "Invalid Wallbox login")

            self.authorization = AuthorizationTokens(**response)
            logger.info("Logged in to EV charger")
        except HTTPError as error:
            raise ConfigurationException(
                "wallbox", "Unable to login to EV charger"
            ) from error

    def _refresh_token(self):
        logger.info("Refreshing access token Wallbox...")
        response = self._post(
            REFRESH_URL.format(host=self.settings.host),
            headers={"Authorization": f"Bearer {self.authorization.refresh_token}"},
            verify=False,
            login=self.login,
        )

        self.authorization.access_token = response["accessToken"]
        logger.info("Refreshed access token Wallbox")



from __future__ import annotations

from functools import lru_cache
from os import environ, listdir, path
from time import localtime, strftime
from typing import Generator

from pydantic import BaseModel, Field, SecretStr

from solaredge2mqtt.logging import LoggingLevelEnum, logger

DOCKER_SECRETS_DIR = "/run/secrets"

SECONDS_PER_DAY = 86400
SECONDS_PER_HOUR = 3600
SECONDS_PER_YEAR = SECONDS_PER_DAY * 365
SECONDS_PER_2_YEARS = SECONDS_PER_YEAR * 2


class ModbusSettings(BaseModel):
    host: str
    port: int = Field(1502)
    timeout: int = Field(1)
    unit: int = Field(1)


class MQTTSettings(BaseModel):
    client_id: str = Field("solaredge2mqtt")
    broker: str
    port: int = Field(1883)
    username: str
    password: SecretStr
    topic_prefix: str = Field("solaredge")


class MonitoringSettings(BaseModel):
    site_id: str = Field(None)
    username: str = Field(None)
    password: SecretStr = Field(None)

    @property
    def is_configured(self) -> bool:
        return all(
            [
                self.site_id is not None,
                self.username is not None,
                self.password is not None,
            ]
        )


class WallboxSettings(BaseModel):
    host: str = Field(None)
    password: SecretStr = Field(None)
    serial: str = Field(None)

    @property
    def is_configured(self) -> bool:
        return all(
            [self.host is not None, self.password is not None, self.serial is not None]
        )


class InfluxDBSettings(BaseModel):
    host: str = Field(None)
    port: int = Field(8086)
    token: SecretStr = Field(None)
    org: str = Field(None)
    bucket: str = Field("solaredge")
    retention: int = Field(SECONDS_PER_2_YEARS)
    retention_raw: int = Field(25)
    timezone: str = Field(strftime("%Z", localtime()))

    @property
    def is_configured(self) -> bool:
        return all(
            [
                self.host is not None,
                self.port is not None,
                self.token is not None,
                self.org is not None,
            ]
        )


class PriceSettings(BaseModel):
    consumption: float = Field(None)
    delivery: float = Field(None)

    @property
    def is_configured(self) -> bool:
        return self.is_consumption_configured or self.is_delivery_configured

    @property
    def is_consumption_configured(self) -> bool:
        return self.consumption is not None

    @property
    def is_delivery_configured(self) -> bool:
        return self.delivery is not None

    @property
    def price_in(self) -> float:
        return self.consumption or 0.0

    @property
    def price_out(self) -> float:
        return self.delivery or 0.0


class LocationSettings(BaseModel):
    latitude: float
    longitude: float

    timezone: str = Field(strftime("%Z", localtime()))


class WeatherSettings(BaseModel):
    api_key: SecretStr | None = Field(None)
    language: str = Field("en")

    @property
    def is_configured(self) -> bool:
        return self.api_key is not None


class ForecastSettings(BaseModel):
    enable: bool = Field(False)
    hyperparametertuning: bool = Field(False)

    @property
    def is_configured(self) -> bool:
        return self.enable


class ServiceSettings(BaseModel):
    interval: int = Field(5)
    logging_level: LoggingLevelEnum = LoggingLevelEnum.INFO

    modbus: ModbusSettings
    mqtt: MQTTSettings

    location: LocationSettings | None = None
    prices: PriceSettings = PriceSettings()

    monitoring: MonitoringSettings | None = None
    wallbox: WallboxSettings | None = None

    influxdb: InfluxDBSettings | None = None

    weather: WeatherSettings | None = None

    forecast: ForecastSettings | None = None

    def __init__(self, **data: dict[str, any]):
        sources = [self._read_environment, self._read_dotenv, self._read_secrets]
        data = self._parse_key_and_values(sources, data)
        super().__init__(**data)

    @property
    def is_location_configured(self) -> bool:
        return self.location is not None

    @property
    def is_prices_configured(self) -> bool:
        return self.prices is not None and self.prices.is_configured

    @property
    def is_monitoring_configured(self) -> bool:
        return self.monitoring is not None and self.monitoring.is_configured

    @property
    def is_wallbox_configured(self) -> bool:
        return self.wallbox is not None and self.wallbox.is_configured

    @property
    def is_influxdb_configured(self) -> bool:
        return self.influxdb is not None and self.influxdb.is_configured

    @property
    def is_weather_configured(self) -> bool:
        is_configured = self.weather is not None and self.weather.is_configured

        if is_configured and not self.is_location_configured:
            logger.warning(
                "Weather settings are configured but location is not configured."
            )
            is_configured = False

        return is_configured

    @property
    def is_forecast_configured(self) -> bool:
        is_configured = self.forecast is not None and self.forecast.is_configured

        if is_configured and not self.is_location_configured:
            logger.warning(
                "Forecast settings are configured but location is not configured."
            )
            is_configured = False

        if is_configured and not self.is_weather_configured:
            logger.warning(
                "Forecast settings are configured but weather is not configured."
            )
            is_configured = False

        return is_configured

    def _parse_key_and_values(
        self, sources: list[callable], data: dict[str, any]
    ) -> dict[str, any]:
        for source in sources:
            for key, value in source():
                key = key.lower().strip()[8:]  # remove prefix
                subkeys = key.split("__")  # get nested structure
                context = data
                for subkey in subkeys[:-1]:
                    if subkey not in context:
                        context[subkey] = {}
                    context = context[subkey]

                context[subkeys[-1]] = (
                    value.strip()
                )  # Missing possibility to set nested json values

        return data

    @classmethod
    def _read_environment(cls) -> Generator[tuple[str, str], any, any]:
        for key, value in environ.items():
            if cls._has_prefix(key):
                yield key, value

    @classmethod
    def _read_secrets(cls) -> Generator[tuple[str, str], any, any]:
        if path.exists(DOCKER_SECRETS_DIR) and path.isdir(DOCKER_SECRETS_DIR):
            for filename in listdir(DOCKER_SECRETS_DIR):
                if cls._has_prefix(filename):
                    with open(
                        path.join(DOCKER_SECRETS_DIR, filename), "r", encoding="utf-8"
                    ) as f:
                        yield filename, f.read()

    @classmethod
    def _read_dotenv(cls) -> Generator[tuple[str, str], any, any]:
        if path.exists(".env"):
            with open(".env", "r", encoding="utf-8") as f:
                for line in f.readlines():
                    line = line.strip()
                    if cls._has_prefix(line) and "=" in line:
                        key, value = line.split("=", 1)
                        yield key, value

    @staticmethod
    def _has_prefix(key: str) -> bool:
        return key.lower().startswith("se2mqtt_")


@lru_cache()
def service_settings() -> ServiceSettings:
    return ServiceSettings()



from json import loads as json_loads
from typing import Callable

from requests import session
from requests.exceptions import ConnectionError as RequestsConnectionError
from requests.exceptions import Timeout, HTTPError

from solaredge2mqtt.logging import logger


class HTTPClient:
    def __init__(self, service: str) -> None:
        self.session = session()
        self.service = service

    def _get(
        self,
        url: str,
        params: dict[str, str | int | float] | None = None,
        headers: dict[str, str] | None = None,
        timeout: int = 5,
        verify: bool = True,
        login: Callable | None = None,
    ) -> dict | None:
        result = None
        try:
            response = self.session.get(
                url, params=params, headers=headers, timeout=timeout, verify=verify
            )

            if response.status_code in (401, 403) and login:
                # Retry with a new login
                login()
                result = self._get(url, headers, timeout, verify)
            else:
                response.raise_for_status()
                result = json_loads(response.content.decode("utf-8"))
        except RequestsConnectionError as error:
            logger.warning(f"Cannot connect to {self.service}: {error}")
        except Timeout as error:
            logger.warning(f"Connection to {self.service} timed out: {error}")

        return result

    def _post(
        self,
        url: str,
        json: dict[str, str] | None = None,
        data: dict[str, str] | None = None,
        headers: dict[str, str] | None = None,
        timeout: int = 5,
        verify: bool = True,
        expect_json: bool = True,
        login: Callable | None = None,
    ) -> dict | None:
        result = None
        try:
            response = self.session.post(
                url,
                headers=headers,
                json=json,
                data=data,
                timeout=timeout,
                verify=verify,
            )

            if response.status_code in (401, 403) and login:
                # Retry with a new login
                login()
                result = self._post(
                    url, json, data, headers, timeout, verify, expect_json
                )
            else:
                response.raise_for_status()
                if expect_json:
                    result = json_loads(response.content.decode("utf-8"))
                else:
                    result = response.content.decode("utf-8")
        except RequestsConnectionError as error:
            logger.warning(f"Cannot connect to {self.service}: {error}")
        except Timeout as error:
            logger.warning(f"Connection to {self.service} timed out: {error}")
        except HTTPError as error:
            logger.warning(f"HTTP error from {self.service}: {error}")
            raise error

        return result
